#include <stdio.h>
#include <stdint.h>
#include <stdlib.h>
#include <string.h>
#include <assert.h>
#include <math.h>
#include <SDL2/SDL.h>
#include <SDL2/SDL_error.h>
#include <SDL2/SDL_video.h>
#include <epoxy/gl.h>

#include <algorithm>
#include <chrono>
#include <memory>
#include <numeric>
#include <random>
#include <vector>
#include <unordered_map>

#include <movit/util.h>

#include "util.h"

#define WIDTH 1280
#define HEIGHT 720
#define WIDTH_BLOCKS (WIDTH/8)
#define WIDTH_BLOCKS_CHROMA (WIDTH/16)
#define HEIGHT_BLOCKS (HEIGHT/8)
#define NUM_BLOCKS (WIDTH_BLOCKS * HEIGHT_BLOCKS)
#define NUM_BLOCKS_CHROMA (WIDTH_BLOCKS_CHROMA * HEIGHT_BLOCKS)

#define NUM_SYMS 256
#define ESCAPE_LIMIT (NUM_SYMS - 1)
#define BLOCKS_PER_STREAM 320

static constexpr uint32_t prob_bits = 12;
static constexpr uint32_t prob_scale = 1 << prob_bits;

unsigned char rgb[WIDTH * HEIGHT * 3];
unsigned char pix_y[WIDTH * HEIGHT];
unsigned char pix_cb[(WIDTH/2) * HEIGHT];
unsigned char pix_cr[(WIDTH/2) * HEIGHT];

struct RansCountSSBO {
	unsigned dist[4 * 256];
	unsigned ransfreq[4 * 256];
};

struct RansDistUBO {
	struct {
		uint32_t x_max, rcp_freq, bias, rcp_shift_and_cmpl_freq;
	} ransdist[4 * 256];
	struct {
		uint32_t val;
		uint32_t padding[3];  // std140 layout.
	} sign_biases[4];
};

using namespace std;
using namespace std::chrono;

void write_varint(int x, FILE *fp)
{
	while (x >= 128) {
		putc((x & 0x7f) | 0x80, fp);
		x >>= 7;
	}
	putc(x, fp);
}

void readpix(unsigned char *ptr, const char *filename)
{
	FILE *fp = fopen(filename, "rb");
	if (fp == nullptr) {
		perror(filename);
		exit(1);
	}

	fseek(fp, 0, SEEK_END);
	long len = ftell(fp);
	assert(len >= WIDTH * HEIGHT * 3);
	fseek(fp, len - WIDTH * HEIGHT * 3, SEEK_SET);

	fread(ptr, 1, WIDTH * HEIGHT * 3, fp);
	fclose(fp);
}

// Should be done on the GPU, of course, but irrelevant for the demonstration.
void convert_ycbcr()
{
	double coeff[3] = { 0.2126, 0.7152, 0.0722 };  // sum = 1.0
	double cb_fac = 1.0 / (coeff[0] + coeff[1] + 1.0f - coeff[2]);  // 0.539
	double cr_fac = 1.0 / (1.0f - coeff[0] + coeff[1] + coeff[2]);  // 0.635 

	unique_ptr<float[]> temp_cb(new float[WIDTH * HEIGHT]);
	unique_ptr<float[]> temp_cr(new float[WIDTH * HEIGHT]);
	for (unsigned yb = 0; yb < HEIGHT; ++yb) {
		for (unsigned xb = 0; xb < WIDTH; ++xb) {
			int r = rgb[((yb * WIDTH) + xb) * 3 + 0];
			int g = rgb[((yb * WIDTH) + xb) * 3 + 1];
			int b = rgb[((yb * WIDTH) + xb) * 3 + 2];
			double y = std::min(std::max(coeff[0] * r + coeff[1] * g + coeff[2] * b, 0.0), 255.0);
			double cb = (b - y) * cb_fac + 128.0;
			double cr = (r - y) * cr_fac + 128.0;
			pix_y[(yb * WIDTH) + xb] = lrint(y);
			temp_cb[(yb * WIDTH) + xb] = cb;
			temp_cr[(yb * WIDTH) + xb] = cr;
		}
	}

	// Simple 4:2:2 subsampling with left convention.
	for (unsigned yb = 0; yb < HEIGHT; ++yb) {
		for (unsigned xb = 0; xb < WIDTH / 2; ++xb) {
			int c0 = yb * WIDTH + std::max(int(xb) * 2 - 1, 0);
			int c1 = yb * WIDTH + xb * 2;
			int c2 = yb * WIDTH + xb * 2 + 1;
			
			double cb = 0.25 * temp_cb[c0] + 0.5 * temp_cb[c1] + 0.25 * temp_cb[c2];
			double cr = 0.25 * temp_cr[c0] + 0.5 * temp_cr[c1] + 0.25 * temp_cr[c2];
			cb = std::min(std::max(cb, 0.0), 255.0);
			cr = std::min(std::max(cr, 0.0), 255.0);
			pix_cb[(yb * WIDTH/2) + xb] = lrint(cb);
			pix_cr[(yb * WIDTH/2) + xb] = lrint(cr);
		}
	}
}

int main(int argc, char **argv)
{
	// Set up an OpenGL context using SDL.
	if (SDL_Init(SDL_INIT_VIDEO) == -1) {
		fprintf(stderr, "SDL_Init failed: %s\n", SDL_GetError());
		exit(1);
	}
	SDL_GL_SetAttribute(SDL_GL_DEPTH_SIZE, 0);
	SDL_GL_SetAttribute(SDL_GL_STENCIL_SIZE, 0);
	SDL_GL_SetAttribute(SDL_GL_DOUBLEBUFFER, 1);
	SDL_GL_SetAttribute(SDL_GL_CONTEXT_PROFILE_MASK, SDL_GL_CONTEXT_PROFILE_CORE);
	SDL_GL_SetAttribute(SDL_GL_CONTEXT_MAJOR_VERSION, 4);
	SDL_GL_SetAttribute(SDL_GL_CONTEXT_MINOR_VERSION, 5);

	SDL_Window *window = SDL_CreateWindow("OpenGL window for unit test",
		SDL_WINDOWPOS_UNDEFINED,
		SDL_WINDOWPOS_UNDEFINED,
		32, 32,
		SDL_WINDOW_OPENGL);
	SDL_GLContext context = SDL_GL_CreateContext(window);
	assert(context != nullptr);

	if (argc >= 2)
		readpix(rgb, argv[1]);
	else
		readpix(rgb, "color.pnm");
	convert_ycbcr();

	// Compile the DCT shader.
	string shader_src = ::read_file("encoder.shader");
	GLuint shader_num = compile_shader(shader_src, GL_COMPUTE_SHADER);
	GLuint glsl_program_num = glCreateProgram();
	glAttachShader(glsl_program_num, shader_num);
	glLinkProgram(glsl_program_num);

	GLint success;
	glGetProgramiv(glsl_program_num, GL_LINK_STATUS, &success);
	if (success == GL_FALSE) {
		GLchar error_log[1024] = {0};
		glGetProgramInfoLog(glsl_program_num, 1024, nullptr, error_log);
		fprintf(stderr, "Error linking program: %s\n", error_log);
		exit(1);
	}

	// Compile the tally shader.
	shader_src = ::read_file("tally.shader");
	shader_num = compile_shader(shader_src, GL_COMPUTE_SHADER);
	GLuint glsl_tally_program_num = glCreateProgram();
	glAttachShader(glsl_tally_program_num, shader_num);
	glLinkProgram(glsl_tally_program_num);

	glGetProgramiv(glsl_tally_program_num, GL_LINK_STATUS, &success);
	if (success == GL_FALSE) {
		GLchar error_log[1024] = {0};
		glGetProgramInfoLog(glsl_tally_program_num, 1024, nullptr, error_log);
		fprintf(stderr, "Error linking program: %s\n", error_log);
		exit(1);
	}

	// Compile the rANS shader.
	shader_src = ::read_file("rans.shader");
	shader_num = compile_shader(shader_src, GL_COMPUTE_SHADER);
	GLuint glsl_rans_program_num = glCreateProgram();
	glAttachShader(glsl_rans_program_num, shader_num);
	glLinkProgram(glsl_rans_program_num);

	glGetProgramiv(glsl_rans_program_num, GL_LINK_STATUS, &success);
	if (success == GL_FALSE) {
		GLchar error_log[1024] = {0};
		glGetProgramInfoLog(glsl_rans_program_num, 1024, nullptr, error_log);
		fprintf(stderr, "Error linking program: %s\n", error_log);
		exit(1);
	}
	check_error();

	// An SSBO for the raw rANS counts.
	GLuint ssbo;
	glGenBuffers(1, &ssbo);
	glBindBuffer(GL_SHADER_STORAGE_BUFFER, ssbo);
	glNamedBufferStorage(ssbo, sizeof(RansCountSSBO), nullptr, GL_MAP_READ_BIT | GL_MAP_PERSISTENT_BIT);
	check_error();

	// UBO for the rANS distributions (copied from an SSBO).
	GLuint dist_ssbo;
	glGenBuffers(1, &dist_ssbo);
	glBindBuffer(GL_SHADER_STORAGE_BUFFER, dist_ssbo);
	glNamedBufferStorage(dist_ssbo, sizeof(RansDistUBO), nullptr, GL_MAP_READ_BIT | GL_MAP_PERSISTENT_BIT);
	check_error();

	GLuint dist_ubo;
	glGenBuffers(1, &dist_ubo);
	glBindBuffer(GL_UNIFORM_BUFFER, dist_ubo);
	glNamedBufferStorage(dist_ubo, sizeof(RansDistUBO), nullptr, GL_MAP_READ_BIT | GL_MAP_PERSISTENT_BIT);
	check_error();

	// SSBOs for the rANS output (data and offsets).
	GLuint output_ssbo;
	glGenBuffers(1, &output_ssbo);
	glBindBuffer(GL_SHADER_STORAGE_BUFFER, output_ssbo);
	glNamedBufferStorage(output_ssbo, 45 * 64 * 1024, nullptr, GL_MAP_READ_BIT | GL_MAP_PERSISTENT_BIT);
	check_error();

	GLuint output_offset_ssbo;
	glGenBuffers(1, &output_offset_ssbo);
	glBindBuffer(GL_SHADER_STORAGE_BUFFER, output_offset_ssbo);
	glNamedBufferStorage(output_offset_ssbo, 45 * 64 * sizeof(uint32_t), nullptr, GL_MAP_READ_BIT | GL_MAP_PERSISTENT_BIT);
	check_error();

	// Bind SSBOs.
	glUseProgram(glsl_program_num);
	glBindBufferBase(GL_SHADER_STORAGE_BUFFER, 9, ssbo);

	glUseProgram(glsl_tally_program_num);
	glBindBufferBase(GL_SHADER_STORAGE_BUFFER, 9, ssbo);
	glBindBufferBase(GL_SHADER_STORAGE_BUFFER, 12, dist_ssbo);

	glUseProgram(glsl_rans_program_num);
	glBindBufferBase(GL_SHADER_STORAGE_BUFFER, 10, output_ssbo);
	glBindBufferBase(GL_SHADER_STORAGE_BUFFER, 11, output_offset_ssbo);
	glBindBufferBase(GL_UNIFORM_BUFFER, 13, dist_ubo);

	glUseProgram(glsl_program_num);
	check_error();

	// Upload luma.
	GLuint y_tex;
	glGenTextures(1, &y_tex);
        glBindTexture(GL_TEXTURE_2D, y_tex);
        glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST);
        glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST);
        glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_REPEAT);
        glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_REPEAT);
        glTexImage2D(GL_TEXTURE_2D, 0, GL_R8UI, WIDTH, HEIGHT, 0, GL_RED_INTEGER, GL_UNSIGNED_BYTE, pix_y);
	check_error();

	// Make destination textures.
	GLuint dc_ac7_tex, ac1_ac6_tex, ac2_ac5_tex;
	for (GLuint *tex : { &dc_ac7_tex, &ac1_ac6_tex, &ac2_ac5_tex }) {
		glGenTextures(1, tex);
		glBindTexture(GL_TEXTURE_2D, *tex);
		glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST);
		glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST);
		glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_REPEAT);
		glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_REPEAT);
		glTexImage2D(GL_TEXTURE_2D, 0, GL_R16UI, WIDTH / 8, HEIGHT, 0, GL_RED_INTEGER, GL_UNSIGNED_SHORT, nullptr);
		check_error();
	}

	GLuint ac3_tex, ac4_tex;
	for (GLuint *tex : { &ac3_tex, &ac4_tex }) {
		glGenTextures(1, tex);
		glBindTexture(GL_TEXTURE_2D, *tex);
		glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST);
		glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST);
		glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_REPEAT);
		glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_REPEAT);
		glTexImage2D(GL_TEXTURE_2D, 0, GL_R8I, WIDTH / 8, HEIGHT, 0, GL_RED_INTEGER, GL_BYTE, nullptr);
		check_error();
	}

	glBindImageTexture(0, dc_ac7_tex, 0, GL_FALSE, 0, GL_READ_WRITE, GL_R16UI);
	glBindImageTexture(1, ac1_ac6_tex, 0, GL_FALSE, 0, GL_READ_WRITE, GL_R16UI);
	glBindImageTexture(2, ac2_ac5_tex, 0, GL_FALSE, 0, GL_READ_WRITE, GL_R16UI);
	glBindImageTexture(3, ac3_tex, 0, GL_FALSE, 0, GL_READ_WRITE, GL_R8I);
	glBindImageTexture(4, ac4_tex, 0, GL_FALSE, 0, GL_READ_WRITE, GL_R8I);
	glBindImageTexture(5, y_tex, 0, GL_FALSE, 0, GL_READ_ONLY, GL_R8UI);
	check_error();

	// Bind uniforms.
	glUseProgram(glsl_program_num);
	GLint dc_ac7_tex_uniform = glGetUniformLocation(glsl_program_num, "dc_ac7_tex");
	GLint ac1_ac6_tex_uniform = glGetUniformLocation(glsl_program_num, "ac1_ac6_tex");
	GLint ac2_ac5_tex_uniform = glGetUniformLocation(glsl_program_num, "ac2_ac5_tex");
	GLint ac3_tex_uniform = glGetUniformLocation(glsl_program_num, "ac3_tex");
	GLint ac4_tex_uniform = glGetUniformLocation(glsl_program_num, "ac4_tex");
	GLint image_tex_uniform = glGetUniformLocation(glsl_program_num, "image_tex");
	glUniform1i(dc_ac7_tex_uniform, 0);
	glUniform1i(ac1_ac6_tex_uniform, 1);
	glUniform1i(ac2_ac5_tex_uniform, 2);
	glUniform1i(ac3_tex_uniform, 3);
	glUniform1i(ac4_tex_uniform, 4);
	glUniform1i(image_tex_uniform, 5);

	glUseProgram(glsl_rans_program_num);
	dc_ac7_tex_uniform = glGetUniformLocation(glsl_rans_program_num, "dc_ac7_tex");
	ac1_ac6_tex_uniform = glGetUniformLocation(glsl_rans_program_num, "ac1_ac6_tex");
	ac2_ac5_tex_uniform = glGetUniformLocation(glsl_rans_program_num, "ac2_ac5_tex");
	ac3_tex_uniform = glGetUniformLocation(glsl_rans_program_num, "ac3_tex");
	ac4_tex_uniform = glGetUniformLocation(glsl_rans_program_num, "ac4_tex");
	image_tex_uniform = glGetUniformLocation(glsl_rans_program_num, "image_tex");
	glUniform1i(dc_ac7_tex_uniform, 0);
	glUniform1i(ac1_ac6_tex_uniform, 1);
	glUniform1i(ac2_ac5_tex_uniform, 2);
	glUniform1i(ac3_tex_uniform, 3);
	glUniform1i(ac4_tex_uniform, 4);

	steady_clock::time_point start = steady_clock::now();
	unsigned num_iterations = 100;
	for (unsigned i = 0; i < num_iterations; ++i) {
		glClearNamedBufferSubData(ssbo, GL_R8, 0, sizeof(RansCountSSBO), GL_RED, GL_UNSIGNED_BYTE, nullptr);
		glUseProgram(glsl_program_num);
		glDispatchCompute(WIDTH_BLOCKS / 16, HEIGHT_BLOCKS, 1);
		glMemoryBarrier(GL_SHADER_STORAGE_BARRIER_BIT);

		glUseProgram(glsl_tally_program_num);
		glDispatchCompute(4, 1, 1);
		glMemoryBarrier(GL_SHADER_STORAGE_BARRIER_BIT);
	
		glCopyNamedBufferSubData(dist_ssbo, dist_ubo, 0, 0, sizeof(RansDistUBO));
		glMemoryBarrier(GL_UNIFORM_BARRIER_BIT);

		glUseProgram(glsl_rans_program_num);
		glDispatchCompute(NUM_BLOCKS / BLOCKS_PER_STREAM, 8, 5);
	}
	check_error();
	glFinish();
	check_error();
	steady_clock::time_point now = steady_clock::now();

#if 0
	printf("%ld bytes + %ld escape bits (%ld) = %ld total bytes\n",
		tot_bytes - extra_bits / 8,
		extra_bits,
		extra_bits / 8,
		tot_bytes);

	printf("\n");
#endif

	printf("Each iteration took %.3f ms.\n", 1e3 * duration<double>(now - start).count() / num_iterations);

	FILE *codedfp = fopen("coded.dat", "wb");
	if (codedfp == nullptr) {
		perror("coded.dat");
		exit(1);
	}

	// Write out the distributions.
	const RansCountSSBO *rans_count = (const RansCountSSBO *)glMapNamedBufferRange(ssbo, 0, sizeof(RansCountSSBO), GL_MAP_READ_BIT | GL_MAP_PERSISTENT_BIT);
	const RansDistUBO *rans_dist = (const RansDistUBO *)glMapNamedBufferRange(dist_ssbo, 0, sizeof(RansDistUBO), GL_MAP_READ_BIT | GL_MAP_PERSISTENT_BIT);
	for (unsigned r = 0; r < 2; ++r) {  // Hack to write fake chroma tables.
		// TODO: rather gamma-k or something
		for (unsigned i = 0; i < 4; ++i) {
			printf("writing table %d\n", i);
			for (unsigned j = 0; j < NUM_SYMS; ++j) {
				printf("%d,%d: freq=%d  x_max=%d, rcp_freq=%08x, bias=%d, rcp_shift=%d, cmpl_freq=%d\n",
					i, j, rans_count->ransfreq[i * 256 + j],
					rans_dist->ransdist[i * 256 + j].x_max,
					rans_dist->ransdist[i * 256 + j].rcp_freq,
					rans_dist->ransdist[i * 256 + j].bias,
					rans_dist->ransdist[i * 256 + j].rcp_shift_and_cmpl_freq & 0xffff,
					rans_dist->ransdist[i * 256 + j].rcp_shift_and_cmpl_freq >> 16);
				write_varint(rans_count->ransfreq[i * 256 + j], codedfp);
			}
		}
	}

	// Write out the actual data.
	// TODO: Do the deduplication.

	const uint32_t *offsets = (const uint32_t *)glMapNamedBufferRange(output_offset_ssbo, 0, 45 * 64 * sizeof(uint32_t), GL_MAP_READ_BIT | GL_MAP_PERSISTENT_BIT);
#if 0
	for (int i = 0; i < 45*64; ++i) {
		printf("%d,%d,%d: %u\n", i / 64, (i / 8) % 8, i % 8, 1024 * (i + 1) - offsets[i]);
	}
#endif

	const uint8_t *data = (const uint8_t *)glMapNamedBufferRange(output_ssbo, 0, 45 * 64 * 1024, GL_MAP_READ_BIT | GL_MAP_PERSISTENT_BIT);

	for (unsigned y = 0; y < 8; ++y) {
		for (unsigned x = 0; x < 8; ++x) {
			for (unsigned int stream_idx = 0; stream_idx < 45; ++stream_idx) {
				const uint8_t *out_end = data + (stream_idx * 64 + y * 8 + x + 1) * 1024;
				const uint8_t *ptr = data + offsets[stream_idx * 64 + y * 8 + x];
				uint32_t num_rans_bytes = out_end - ptr;
#if 0
				if (num_rans_bytes == last_block.size() &&
				    memcmp(last_block.data(), ptr, last_block.size()) == 0) {
					write_varint(0, codedfp);
					clear();
					return 1;
				} else {
					last_block = string((const char *)ptr, num_rans_bytes);
				}
#endif

				write_varint(num_rans_bytes, codedfp);
				fwrite(ptr, 1, num_rans_bytes, codedfp);
			}
		}
	}
	fclose(codedfp);
}
